""""""
from collections import OrderedDict
import datetime as dt
from distutils.dir_util import copy_tree
import functools
import numpy as np
import os
import pandas as pd
import pandas.io.formats.excel as pde
import time
import yfinance as yf


def constituents(start_date, end_date, lag, data_path):
    """For each symbol, extracts the date ranges of index inclusion between the start and end date from the histfile.

    Is called within the `import_data()` method of :class:`ndxtest.backtest.BackTest`. `edr` refers to the extended date
    range including the lag.

    :param datetime.datetime start_date: The start date of the backtest period.
    :param datetime.datetime end_date: The end date of the backtest period.
    :param datetime.timedelta lag: A trailing timedelta added to the start_date. Necessary for calculating lagging indicators such as moving averages.
    :param str data_path: Has to represent the absolute path to where the lib folder is stored.

    :returns: A nested dictionary as follows: {symbol: {`dr`: pandas.daterange, `edr`: pandas.daterange}, ...}
    :rtype: dict
    """

    hist = pd.read_excel(data_path + 'lib\\^HIST.xlsx', index_col='date', parse_dates=[0])

    dr = pd.date_range(start_date, end_date)
    edr = pd.date_range(start_date - lag, end_date)
    hist = hist.loc[hist.index.intersection(dr)]

    tickers = []
    for row in hist.symbols.values:
        for s in row.split(','):
            if s not in tickers:
                tickers += [s]

    cons = {k: None for k in tickers}

    added = [s for s in hist.added.values if str(s) != 'nan']
    removed = [s for s in hist.removed.values if str(s) != 'nan']

    multiple = list(set(added).intersection(set(removed)))
    added = [s for s in added if s not in multiple]
    removed = [s for s in removed if s not in multiple]
    whole_period = [s for s in tickers if s not in added and s not in removed and s not in multiple]

    for s in multiple:
        df = hist.loc[(hist['added'] == s) | (hist['removed'] == s)]
        if len(df.index) > 2:
            print(f'Warning: {s} had > 2 index inclusions/removals during time period.')

        if df.loc[df.index[0]].added == s:
            cons[s] = {'dr': pd.date_range(df.index[0], df.index[1]),
                       'edr': pd.date_range(df.index[0] - lag, df.index[1])}
        if df.loc[df.index[0]].removed == s:
            cons[f'{s}'] = {'dr': pd.date_range(start_date, df.index[0]), 'edr': pd.date_range(start_date - lag, df.index[0])}
            cons[f'{s}*'] = {'dr': pd.date_range(df.index[1], end_date), 'edr': pd.date_range(df.index[1] - lag, end_date)}

    for s in whole_period:
        cons[s] = {'dr': dr,
                   'edr': edr}
    for s in added:
        cons[s] = {'dr': pd.date_range(hist.loc[hist.added == s].index[0], end_date),
                   'edr': pd.date_range(hist.loc[hist.added == s].index[0] - lag, end_date)}
        if s in removed:
            cons[s] = {'dr': pd.date_range(hist.loc[hist.added == s].index[0], hist.loc[hist.removed == s].index[0]),
                       'edr': pd.date_range(hist.loc[hist.added == s].index[0] - lag, hist.loc[hist.removed == s].index[0])}
            removed.remove(s)
    for s in removed:
        cons[s] = {'dr': pd.date_range(start_date, hist.loc[hist.removed == s].index[0]),
                   'edr': pd.date_range(start_date - lag, hist.loc[hist.removed == s].index[0])}
    return cons


def connect(data_path, gspc_strict=False, hist_strict=False):
    """Performs checks on the provided `data_path`.

    The `data_path` has to be either the parent directory containing the `data` folder or
    the directory of the `data` folder itself. The `data` folder has to contain the `lib`
    folder.

    :param str data_path: The `data` directory or its parent directory.
    :param bool, default=False, gspc_strict: If True, '^GSPC.csv' has to be in the `lib` folder.
    :param bool, default=False, hist_strict: If True, '^HIST.xlsx' has to be in the `lib` folder.

    :raises: ValueError if data_path was incorrect.

    :returns: The data_path if it has passed the checks.
    :rtype: str
    """

    if not isinstance(data_path, str):
        raise TypeError('data_path was not of type str')

    if data_path.endswith('\\'):
        pass
    else:
        data_path += '\\'

    if data_path.endswith('data\\'):
        pass
    elif 'data' in os.listdir(data_path):
        data_path += 'data\\'
    else:
        raise ValueError("The provided path does not end with and does not contain data\\.")

    if 'lib' not in os.listdir(data_path):
        raise ValueError("data\\ found, but folder lib\\ in data\\.")

    if gspc_strict:
        if '^GSPC.csv' not in os.listdir(data_path + 'lib\\'):
            raise FileNotFoundError("data\\lib\\ found, but '^GSPC.csv' is missing.")

    if hist_strict:
        if '^HIST.xlsx' not in os.listdir(data_path + 'lib\\'):
            raise FileNotFoundError("data\\lib\\ found, but '^HIST.xlsx' is missing.")

    return data_path


def timeit_decorator(func):
    """This decorator prints the time needed for function execution to the console.

    :param function func: Any function.
    """
    @functools.wraps(func)
    def wrapper_func(*args, **kwargs):
        first = f"Finished {func.__name__}..."
        j = 40 - len(func.__name__)
        indent = ''
        for i in range(0, j):
            indent += ' '
        t0 = time.time()
        func(*args, **kwargs)
        t1 = time.time()
        delta = (t1 - t0).__round__(2)
        print(first + f"{indent}...{delta} sec elapsed")
        return None
    return wrapper_func


class Portfolio:
    """The Portfolio class simulates a portfolio. It is a code container not meaningful for users of ndxtest.

    Instances of the BackTest class instantiate a portfolio from this class. Within the .run_backtest() method of the
    BackTest instance the portfolio and its methods help to structure the codebase and make it more readable.
    The Portfolio class may be shifted to ndxtest.core.utils in future releases.

    :ivar int max_positions: Maximum number of positions (slots). Initial value: param 'max_positions' of __init__
    :ivar float cash: Available unrestricted cash in the balance. Initial value: param 'initial_equity' of __init__
    :ivar float commission: Commission paid upon entering/exiting positions (e.g. 0.01 -> 1%). Initial value: param 'commission' of __init__
    :ivar float invested_capital: Current market value of the invested capital. Initial value: 0
    :ivar float commission_paid: Cumulative commissions paid during the backtest. Initial value: 0
    :ivar int number_of_trades: Cumulative number of trades executed. Is incremented when a position is entered but not when closed/covered. Initial value: 0
    :ivar dict cash_from_short_positions: The 'restricted' cash that is part of the balance as a result of entering short positions. Initial value: {}
    :ivar dict long_positions: Contains information about the current long positions. Initial value: {}
    :ivar dict short_positions: Contains information about the current short positions. Initial value: {}
    :ivar OrderedDict logdict:  Contains information about closed positions. Used to generate a report on the backtest. Initial value: OrderedDict()
    """

    def __init__(self, max_positions, initial_equity, commission):
        """Constructor method.

        :param int max_positions: Maximum number of positions/slots of the portfolio.
        :param float initial_equity: The initial capital to start with.
        :param float commission: Commission paid upon entering/exiting positions (e.g. 0.01 -> 1%).

        """

        self.max_positions = max_positions
        self.cash = initial_equity
        self.commission = commission
        self.invested_capital, self.commission_paid, self.number_of_trades = 0, 0, 0
        self.cash_from_short_positions = {}
        self.long_positions, self.short_positions = {}, {}
        self.logdict = OrderedDict()

    def positions(self):
        """
        :returns: The union of ticker symbols that currently have open long or short positions.
        :rtype: set
        """
        return set(list(self.long_positions.keys()) + list(self.short_positions.keys()))

    def number_of_positions(self):
        """
        :returns: The number of currently filled positions.
        :rtype: int
        """

        return len(self.positions())

    def free_slot(self):
        """
        :returns: True when there is a free slot in the portfolio or False if there is none.
        :rtype: bool
        """

        return True if self.number_of_positions() < self.max_positions else False

    def current_exposure(self):
        """
        :returns: A dict with some info on current positions.
        :rtype: dict
        """

        return {'n_long': len(self.long_positions.keys()),
                'n_short': len(self.short_positions.keys()),
                'n_free': self.max_positions - self.number_of_positions(),
                'net_exposure': len(self.long_positions.keys()) - len(self.short_positions.keys())}

    def restricted_cash(self):
        """Returns the amount of 'restricted' cash, resulting from entering short positions. This cash is not considered
        for calculating the position size when entering new positions.

        :returns: The amount of 'restricted' cash.
        :rtype: float
        """

        return sum(self.cash_from_short_positions.values())

    def calculate_nshares(self, entry_price):
        """Calculates the number of shares to buy/sell based on free slots, share price and cash balance.

        :returns: Number of shares to buy/sell.
        :rtype: int
        """

        return (((self.cash - self.restricted_cash()) /
                 (self.max_positions - self.number_of_positions())) / entry_price).__floor__()

    def describe(self, date):
        """Prints a summary of some Portfolio stats to the console.

        :returns: None
        :rtype: NoneType
        """

        print(f'Stats of {self} as of {date}:')
        print(f'Number of Positions: {self.number_of_positions()}. '
              f'Long: {list(self.long_positions.keys())}, '
              f'Short: {list(self.short_positions.keys())}')
        print(f'Cash: {self.cash.__floor__()}. Cash from shorts: {self.restricted_cash().__floor__()}. '
              f'Invested Capital: {self.invested_capital.__floor__()}.')
        print(f'Total Market Value: {(self.invested_capital + self.cash + self.restricted_cash()).__floor__()}')
        print('\n')

    def create_log_entry(self, entry_data=None, exit_data=None):
        """Creates a record in the instance variable `logdict` upon entering or exiting positions.

        :param dict, default=None entry_data: Contains some data connected with the entry into the position.
        :param dict, default=None exit_data: Contains some data connected with the exit from the position.

        :returns: A dict with aggregated information about the position. Used to create a log record.
        :rtype: dict
        """

        entry_comm = abs(entry_data['entry_price'] * entry_data['nshares'] * self.commission)
        duration, exit_comm, pnl, p_pnl = 0, 0, 0, 0
        if exit_data:
            duration = (exit_data['exit_date'] - entry_data['entry_date']).days
            exit_comm = abs(exit_data['exit_price'] * entry_data['nshares'] * self.commission)
            pnl = (exit_data['exit_price'] - entry_data['entry_price']) * entry_data['nshares']
            p_pnl = ((pnl / (entry_data['entry_price'] * entry_data['nshares'])) * 100)
        d = {'id': entry_data['id'],
             'entry_date': entry_data['entry_date'],
             'entry_weekday': entry_data['entry_date'].isoweekday(),
             'exit_date': exit_data['exit_date'] if exit_data else 0,
             'exit_weekday': exit_data['exit_date'].isoweekday() if exit_data else 0,
             'duration': duration,
             'symbol': entry_data['symbol'],
             'long': 1 if entry_data['signal'] == 1 else 0,
             'short': 1 if entry_data['signal'] == -1 else 0,
             'nshares': entry_data['nshares'],
             'entry_price': entry_data['entry_price'],
             'entry_score': entry_data['entry_score'],
             'entry_value': entry_data['entry_price'] * entry_data['nshares'],
             'exit_price': exit_data['exit_price'] if exit_data else 0,
             'exit_score': exit_data['exit_score'] if exit_data else 0,
             'exit_value': exit_data['exit_price'] * entry_data['nshares'] if exit_data else 0,
             'entry_comm': -entry_comm if not exit_data else 0,
             'exit_comm': -exit_comm,
             '$change': pnl,
             '%change': p_pnl,
             'profitable': 1 if pnl > 0 else 0,
             'p.positions': f'l: {len(list(self.long_positions.keys()))} s: {len(list(self.short_positions.keys()))}',
             'p.cash': self.cash,
             'p.res_cash': self.restricted_cash(),
             'p.invested_capital': self.invested_capital,
             'market_value': (self.invested_capital + self.cash + self.restricted_cash())}
        return d

    def create_log_df(self):
        """
        :returns: A pd.DataFrame, created from the instance variable `logdict`.
        :rtype: pd.DataFrame
        """

        return pd.DataFrame.from_dict(data=self.logdict, orient='index').round(2)

    def long(self, data):  # closing short position has priority
        """Creates a long, or closes a short position in the portfolio. Closing an existing short position has precedence
        over creating a new long position.

        :param dict data: Contains information regarding the entered/covered position.

        :returns: None
        :rtype: NoneType
        """

        if data['symbol'] in self.short_positions.keys():
            nshares = self.short_positions[data['symbol']]['nshares']
            entry_value = self.cash_from_short_positions[data['symbol']]
            current_value = -nshares * data['exit_price']

            self.cash += entry_value - current_value  # subtracting the P/L on the cash side
            self.cash -= abs(self.commission * nshares * data['exit_price'])
            self.commission_paid += abs(self.commission * nshares * data['exit_price'])
            self.invested_capital -= (nshares * data['exit_price']) - (entry_value - current_value)
            del self.cash_from_short_positions[data['symbol']]

            self.logdict[f"{str(data['exit_date'])[:10]}_{data['symbol']}"] = \
                self.create_log_entry(entry_data=self.short_positions[data['symbol']], exit_data=data)

            del self.short_positions[data['symbol']]
        else:  # entering a new long position
            if self.free_slot():
                self.number_of_trades += 1
                data['id'] = self.number_of_trades
                self.long_positions[data['symbol']] = data

                self.cash -= (data['nshares'] * data['entry_price']) + (
                        self.commission * data['nshares'] * data['entry_price'])
                self.commission_paid += self.commission * data['nshares'] * data['entry_price']
                self.invested_capital += data['nshares'] * data['entry_price']
                self.logdict[f"{str(data['entry_date'])[:10]}_{data['symbol']}"] = \
                    self.create_log_entry(entry_data=self.long_positions[data['symbol']])

    def short(self, data):  # closing long position has priority
        """Creates a short, or closes a long position in the portfolio. Closing an existing long position has precedence
        over creating a new short position.

        :param dict data: Contains information regarding the entered/closed position.

        :returns: None
        :rtype: NoneType
        """

        if data['symbol'] in self.long_positions.keys():
            entry_price = self.long_positions[data['symbol']]['entry_price']
            nshares = self.long_positions[data['symbol']]['nshares']

            self.cash += (nshares * data['exit_price']) - (self.commission * nshares * data['exit_price'])
            self.commission_paid += self.commission * nshares * data['exit_price']
            self.invested_capital -= nshares * (entry_price - data['exit_price'])  # subtracting the P/L
            self.invested_capital -= nshares * data['exit_price']
            self.logdict[f"{str(data['exit_date'])[:10]}_{data['symbol']}"] = \
                self.create_log_entry(entry_data=self.long_positions[data['symbol']], exit_data=data)
            del self.long_positions[data['symbol']]
        else:  # entering a new short position
            if self.free_slot():
                self.number_of_trades += 1
                data['id'] = self.number_of_trades
                self.short_positions[data['symbol']] = data

                self.cash -= abs(self.commission * data['nshares'] * data['entry_price'])
                self.cash_from_short_positions[data['symbol']] = -data['nshares'] * data['entry_price']
                self.commission_paid += self.commission * -data['nshares'] * data['entry_price']
                self.invested_capital += data['nshares'] * data['entry_price']

                self.create_log_entry(entry_data=self.short_positions[data['symbol']])


class LibManager:
    """This class is the main tool for maintaining and updating the data library.

    :ivar str data_path: The path to the data.
    :ivar list symbols_in_lib: A list of all symbols found in \\lib.
    """

    def __init__(self, data_path):
        """Constructor method.

        :param str data_path: Has to represent directory to the `data` folder, or the `data` folders parent directory.
        """

        self.data_path = connect(data_path)
        self.symbols_in_lib = [symbol[:-4] for symbol in os.listdir(self.data_path + 'lib\\') if symbol.endswith('.csv')]

    def download_batch(self, symbol_list: list, period='5y'):
        """Downloads price data for new symbols from https://finance.yahoo.com/ and performs some formatting.

        Creates a new folder data\\downloaded_YYYY-MM-DD\\ where downloaded .csv files are saved.

        :param list symbol_list: Valid ticker symbols to download. For example: ['AAPL', 'AMZN', 'META']
        :param str, default='5y' period: Time period of data to download. Valid values are: '1d', '5d', '1mo', '3mo', '6mo',
            '1y', '2y', '5y', '10y', 'ytd' and 'max'. It is recommended to download at least 2 years of price data
            when a symbol is added to the index for the first time. This is important for calculation of indicators
            like 200 period moving averages that require a lot of trailing price data.

        :returns: None
        :rtype: NoneType
        """

        if not isinstance(symbol_list, list):
            raise TypeError("Parameter `symbol_list` must be of type `list`.")

        if not isinstance(period, str):
            raise TypeError("Parameter `period` must be of type `str`.")

        if period not in ['1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y', '10y', 'ytd', 'max']:
            raise ValueError("Parameter `period` must be either '1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y', '10y', 'ytd' or 'max'.")

        if f'downloaded_{str(dt.datetime.now())[:10]}' in os.listdir(self.data_path):
            pass
        else:
            os.mkdir(self.data_path + f'downloaded_{str(dt.datetime.now())[:10]}\\')

        for symbol in symbol_list:
            try:
                data = yf.Ticker(symbol)
                df = data.history(period=period)
                df.insert(0, 'date', df.index)
                df.set_index('date', inplace=True)
                try:
                    df.columns = ['open', 'high', 'low', 'close', 'volume', 'dividends', 'stock_splits']
                except ValueError:
                    pass
                df.insert(0, 'symbol', symbol)
                df['open'] = np.float32(df['open'])
                df['high'] = np.float32(df['high'])
                df['low'] = np.float32(df['low'])
                df['close'] = np.float32(df['close'])

                if 0 in df[['open', 'high', 'low', 'close']].values:
                    print(f'Warning: {symbol} contained missing values in price data.')

                print(f"Downloaded {len(df.index)} records for {symbol}.")

                if len(df.index) > 1:
                    df.to_csv(self.data_path + f'downloaded_{str(dt.datetime.now())[:10]}\\' + f'{symbol}.csv')

            except KeyError or ValueError or AttributeError:
                print(f'Error processing {symbol}... (symbol may be delisted or a name change may have occurred)')

    def lib_update(self, index_symbol='^GSPC', period='3mo', period_first_download='5y', new_entries=0, symbols=None):
        """This function updates all `active` symbols in data\\lib\\.

        Active symbols are read from the from the `histfile`. As of now, the histfile has to be updated manually
        by checking for updates to the index on 'https://www.spglobal.com/spdji/en/indices/equity/sp-500/#news-research'.

        Before updating the contents of data\\lib\\, a backup folder: data\\lib_backup_YYYY-MM-DD\\ is created.
        Missing historic price data between the last update (newest record in ^GSPC.csv) and  is downloaded from yahoo finance
        with the help of the the yfinance package. If a symbol had a stock split between the last and the current update,
        the existing data is processed accordingly before appending the new records.

        :param str, default='^GSPC' index_symbol: '^GSPC' is the ticker symbol of the S&P 500 index on https://finance.yahoo.com/
        :param str, default='3mo' period: The period for which new price data is downloaded, if the last update is more than 3 months back, increase
            accordingly. Valid periods are '1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y', '10y', 'ytd' and 'max'. Be
            aware that downloading large amounts of data for many symbols can lead to restricted access to the yahoo
            finance API on the way. In this case, the process of updating the library will take some time longer.
        :param str, default='5y' period_first_download: Like `period` but for symbols with no existing .csv file.
            Usually, symbols that were newly added to the index
        :param int, default=0 new_entries: The update_lib function only updates symbols that are currently included in the index, i.e. all symbol listed
            in the newest record in the `histfile`. If new entries are added to the histfile BEFORE running
            update_lib, the last row in the histfile will not contain all symbols that need updating. If this is the case,
            increase `new_entries` by 1 for each entry that was added.
        :param list or str, default=None symbols: Here, you can provide a list of symbols if your intention is to only update price data for specific symbols.
            E.g. ['AAPL', 'AMZN'].

        :returns: None
        :rtype: NoneType

        For more information on how to maintain the data library and histfile for the ndxtest package, please refer to
        the user manual.
        """

        self.data_path = connect(self.data_path, gspc_strict=True, hist_strict=True)

        from_dir = self.data_path + 'lib\\'
        to_dir = self.data_path + f'\\lib_backup_{str(dt.datetime.now())[:10]}\\'
        copy_tree(from_dir, to_dir)  # a safety backup of the lib folder is generated prior to updating

        if symbols is None:
            hist = pd.read_excel(self.data_path + 'lib\\^HIST.xlsx')
            activesymbols = sorted(list(set(','.join(hist.symbols.values[-(new_entries + 1):]).split(','))))
            # all symbols within 'last_rows' of the histfile are updated
        else:
            if isinstance(symbols, list):
                activesymbols = symbols
            elif isinstance(symbols, str):
                activesymbols = [symbols]
            else:
                raise TypeError("Parameter `symbols` must be of type `list` or `str`.")

        activesymbols.insert(0, index_symbol)
        approved = False
        breaker = False
        reference_index = None

        for symbol in activesymbols:

            if symbol not in self.symbols_in_lib:  # the symbol has no .csv file in //lib
                print(f"{symbol} not found in lib, downloading {period_first_download}'s of price data...")
                try:
                    data = yf.Ticker(symbol)
                    df = data.history(period=period_first_download)
                    df.insert(0, 'date', df.index)
                    df.set_index('date', inplace=True)
                    try:
                        df.columns = ['open', 'high', 'low', 'close', 'volume', 'dividends', 'stock_splits']
                    except ValueError:
                        pass
                    df.insert(0, 'symbol', symbol)
                    df['open'] = np.float32(df['open'])
                    df['high'] = np.float32(df['high'])
                    df['low'] = np.float32(df['low'])
                    df['close'] = np.float32(df['close'])

                    if 0 in df[['open', 'high', 'low', 'close']].values:
                        print(f'Warning: {symbol} contained missing values in price data.')

                    print(f"{len(df.index)} records downloaded for {symbol}...")

                    if len(df.index) > 1:
                        df.to_csv(self.data_path + f'lib\\{symbol}.csv')

                except KeyError or ValueError or AttributeError:
                    print(f'Error processing {symbol}... (symbol may be delisted or a name change may have occurred)')

            else:  # the symbol already has a .csv file in //lib
                main = pd.read_csv(self.data_path + f'lib\\{symbol}.csv', index_col='date', parse_dates=[0])
                time.sleep(.3)
                try:
                    data = yf.Ticker(symbol)
                    df = data.history(period=period)
                    df.insert(0, 'date', df.index)
                    df.set_index('date', inplace=True)

                    try:
                        df.columns = ['open', 'high', 'low', 'close', 'volume', 'dividends', 'stock_splits']
                    except ValueError:
                        print(f'Error processing {symbol}... (symbol may be delisted or a name change may have occurred)')

                    df.insert(0, 'symbol', symbol)

                    if 0 in df[['open', 'high', 'low', 'close']].values:
                        print(f'Warning: {symbol} contained missing values in price data.')
                        approved = False

                    diff = df.index.difference(main.index)
                    df = df.loc[diff]

                    if any(df.stock_splits.values[:]):
                        s = float(df.loc[df.stock_splits.values[:] != 0, 'stock_splits'])  # s = split_factor
                        print(f'Processed split by factor of {s} for {symbol}.')
                        main = main.divide([None, s, s, s, s, 1 / s, s, 1])
                        main['symbol'] = symbol

                    while not approved:
                        if reference_index is None:
                            print(f'{index_symbol} is used as a reference.')
                            print(f'The last updated was performed on {main.index[-1]}.')
                            print(f'Downloaded data contains {len(df.index)} new records.')
                            print(f'New records: {list(df.index)}.')
                            print(
                                f'Caution, do only perform this updating routine on weekends or after US market close.')
                        if input(f'Proceed appending new records to all files? Y/N:').upper() == 'Y':
                            approved = True
                            reference_index = df.index
                        else:
                            breaker = True
                            break

                    if breaker:
                        print('Process terminated by user.')
                        break

                    if reference_index.identical(df.index):
                        print(f'Appending {len(df.index)} records to {symbol}...')
                        pd.concat([main, df]).to_csv(f'data\\lib\\{symbol}.csv')
                    else:
                        print(f'Downloaded data for {symbol} contained following reference_index:')
                        print(df.index)
                        print(f'This does not match the reference index:')
                        print(reference_index)
                        if input(f'Append nevertheless? Y/N:').upper() == 'Y':
                            print(f'Appending {len(df.index)} records to {symbol}...')
                            pd.concat([main, df]).to_csv(f'data\\lib\\{symbol}.csv')
                        if input(f'Continue with updating routine? Y/N:').upper() == 'N':
                            breaker = True

                except KeyError or ValueError or AttributeError:
                    print(f'Error processing {symbol}... (symbol may be delisted or a name change may have occurred)')
        return None

    def lib_rename_symbol(self, old: str, new: str):
        """Renames a symbol in the library and the `histfile`.

        :param str old: Old name of the symbol, e.g. 'FB'
        :param str new: New name of the symbol, e.g. 'META'

        :returns: None
        :rtype: NoneType
        """

        if not isinstance(old, str) and isinstance(new, str):
            raise TypeError("Parameters `old` and `new` must be of type `str`.")

        self.data_path = connect(self.data_path, hist_strict=True)

        # renaming in the histfile
        pde.ExcelFormatter.header_style = None
        hist = pd.read_excel(self.data_path + 'lib\\^HIST.xlsx', index_col='date', parse_dates=[0])
        data = []
        for row in hist['symbols']:
            row = row.split(',')
            if old in row:
                i = row.index(old)
                new_row = row[:i] + [new] + row[i + 1:]
                data.append(','.join(sorted(new_row)))
            else:
                data.append(','.join(sorted(row)))
        hist['symbols'] = data

        hist['added'].replace(old, new, inplace=True)
        hist['removed'].replace(old, new, inplace=True)

        hist.to_excel(self.data_path + 'lib\\^HIST.xlsx')

        # renaming in the library
        file = pd.read_csv(self.data_path + f'lib\\{old}.csv', index_col='date', parse_dates=[0])
        file['symbol'] = new
        file.to_csv(f'data\\lib\\{new}.csv')
        if new != old:
            os.remove(self.data_path + f'lib\\{old}.csv')
        return None

    def histfile_new_entry(self, action: str, symbol: str, date: str or dt.datetime):
        """Creates a new entry in the `histfile` with a symbol and the date of addition to/removal from the index.

        :param str action: Accepts 'add' for adding a symbol, or 'remove' for removing a symbol.
        :param str symbol: Name of the new symbol, e.g. 'TSLA'.
        :param str or datetime.datetime date: The date on which the symbol was added to/removed from the index. If a str is provided it must be 'YYYY-MM-DD' formatted.

        :returns: None
        :rtype: NoneType

        For more information on how to maintain the histfile please refer to the user manual.
        """

        self.data_path = connect(self.data_path, hist_strict=True)

        if not isinstance(action, str):
            raise TypeError("Argument `action` must be of type str.")

        if (action != 'add') and (action != 'remove'):
            raise ValueError("Argument `action` must be either 'add' or 'remove'")

        if isinstance(date, str):
            if ('-' not in date) or (len(date) != 10):
                raise ValueError("Argument `date` has to be 'YYYY-MM-DD' formatted if type str is provided.")
            date = dt.datetime.strptime(date, '%Y-%m-%d')
        elif isinstance(date, dt.datetime):
            pass
        else:
            raise TypeError("date_added must be a string formatted 'YYYY-MM-DD' or a datetime.datetime object!")

        pde.ExcelFormatter.header_style = None
        hist = pd.read_excel(self.data_path + 'lib\\^HIST.xlsx', index_col='date', parse_dates=[0])

        last_row = hist.symbols.values[-1:][0].split(',')

        if action == 'add':
            last_row.append(symbol)
            new_row = ','.join(sorted(last_row))
        else:
            last_row.remove(symbol)
            new_row = ','.join(sorted(last_row))  # will create an Error if symbol not in last row

        new_entry = pd.DataFrame(data={'added': symbol if action == 'add' else '',
                                       'removed': symbol if action == 'remove' else '',
                                       'symbols': new_row},
                                 index=[date])

        hist = pd.concat([hist, new_entry])
        hist.index.name = 'date'
        hist.to_excel(self.data_path + 'lib\\^HIST.xlsx')
        return None





"""This is the main module of the ndxtest package containing :class:`ndxtest.backtest.BackTest` and
:class:`ndxtest.backtest.Strategy`.
"""

import os
import datetime as dt
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import OrderedDict, defaultdict
from ndxtest.utils import Portfolio, constituents, connect, timeit_decorator
from fpdf import FPDF


class Strategy:
    """Class to build strategies. Strategy objects are provided as a parameter to the `generate_signals` method of
    :class:`ndxtest.backtest.BackTest`.

    :ivar dict data: A dictionary containing price data provided by :class:`ndxtest.backtest.BackTest`. Initial value: None
    :ivar pd.Dataframe index: A a pd.Dataframe containing price data of the reference index. Initial value: None
    :ivar dict condition_sets: A dict of lists with conditions for entering and exiting positions. Initial value: {...}

    For information on how to use this class please refer to the user manual.
    """

    def __init__(self, data=None, index=None):
        """Constructor Method.

        :param dict data: The is data provided when the strategy is provided to the `generate_signals` method of :class:`ndxtest.backtest.BackTest`. Initial value: None
        :param pd.DataFrame index: Is provided when the strategy is provided to the `generate_signals` method of :class:`ndxtest.backtest.BackTest`. Initial value: None
        """

        self.data = data
        self.index = index
        # conditions for entering (en) and exiting (ex) long (l) and short (s) trades
        self.condition_sets = {'enl': [],
                               'exl': [],
                               'ens': [],
                               'exs': []}

    def enter_long_if(self, day, condition, use_index=False):
        """Adding an entry condition for long positions.

        :param int day: Day (relative to day 0 of signal completion) on which to check for the condition to be fulfilled.
        :param function condition: A function of x where x is a `pd.DataFrame` that returns a boolean `pd.Series` with the same index.
        :param bool, default=False use_index: If True, is being applied to the symbol df, else to the reference index df.

        For information on how conditions can be structured please refer to the user manual.

        :returns: None
        :rtype: NoneType
        """
        self.condition_sets['enl'].append((day, condition, use_index))

    def exit_long_if(self, day, condition, use_index=False):
        """Adding an exit condition for long positions.

        :param int day: Day (relative to day 0 of signal completion) on which to check for the condition to be fulfilled.
        :param function condition: A function of x where x is a `pd.DataFrame` that returns a boolean `pd.Series` with the same index.
        :param bool, default=False use_index: If True, is being applied to the symbol df, else to the reference index df.

        For information on how conditions can be structured please refer to the user manual.

        :returns: None
        :rtype: NoneType
        """
        self.condition_sets['exl'].append((day, condition, use_index))

    def enter_short_if(self, day, condition, use_index=False):
        """Adding an entry condition for short positions.

        :param int day: Day (relative to day 0 of signal completion) on which to check for the condition to be fulfilled.
        :param function condition: A function of x where x is a `pd.DataFrame` that returns a boolean `pd.Series` with the same index.
        :param bool, default=False use_index: If True, is being applied to the symbol df, else to the reference index df.

        For information on how conditions can be structured please refer to the user manual.

        :returns: None
        :rtype: NoneType
        """
        self.condition_sets['ens'].append((day, condition, use_index))

    def exit_short_if(self, day, condition, use_index=False):
        """Adding an exit condition for short positions.

        :param int day: Day (relative to day 0 of signal completion) on which to check for the condition to be fulfilled.
        :param function condition: A function of x where x is a `pd.DataFrame` that returns a boolean `pd.Series` with the same index.
        :param bool, default=False use_index: If True, is being applied to the symbol df, else to the reference index df.

        For information on how conditions can be structured please refer to the user manual.

        :returns: None
        :rtype: NoneType
        """
        self.condition_sets['exs'].append((day, condition, use_index))

    def generate_signals(self):
        """Searches in data for signals based on all entry and exit conditions provided to this strategy.

        Discovered signals are annotated as columns 'entry_signals' and 'exit_signals' to each symbol df.
        In the 'entry_signals' column: 0 = no signal (do nothing), 1 = enter long, -1 = enter short.
        In the 'exit_signals' column: 0 = no signal (do nothing), -1 = exit long, 1 = exit short.

        :returns: The price data dict, with signals appended in the described columns.
        :rtype: dict

        For additional information please refer to the user manual.
        """

        if self.data is None or self.index is None:
            print('No data has been passed to this Strategy instance as self.data and/or self.index is None')
            return None

        for symbol, df in self.data.items():  # scanning for entry and exit conditions in data

            self.data[symbol]['score'] = 0
            self.data[symbol]['enl_signals'] = 0
            self.data[symbol]['exl_signals'] = 0
            self.data[symbol]['ens_signals'] = 0
            self.data[symbol]['exs_signals'] = 0

            # intermediate 'index_df' contains the same records as df
            index_df = self.index.loc[df.index.intersection(self.index.index)]

            # processing all signals...
            for k, v in self.condition_sets.items():

                boolean_array = pd.DataFrame()

                for i, condition in enumerate(v):
                    days, func, use_index = condition
                    if use_index:
                        boolean_array[i] = func(index_df).shift(-days)
                    else:
                        boolean_array[i] = func(df).shift(-days)

                signals = boolean_array.loc[boolean_array.T.all()].index.values
                self.data[symbol].loc[signals, f'{k}_signals'] = 1

        return self.data


class BackTest:
    """:class:`ndxtest.backtest.BackTest` runs backtests of trading strategies on an index level. Please refer
    to the user manual for examples how to use this class. The technical documentation focuses on listing
    instance variables and methods.

    :class:`ndxtest.backtest.BackTest` has more instance variables than listed below. The omitted variables are
    of no significance to users of ndxtest.

    :ivar str data_path: Absolute path to the data folder. Initial value: user input (__init__)
    :ivar list data_path_symbols: List of all symbols found in data\\lib. Initial value: []
    :ivar dict input_data: Dict containing the price data with ticker symbols as keys. Initial value: {}
    :ivar dict data: Copy of `input_data` with trading signals added to price data. Initial value: None
    :ivar pd.DataFrame input_index: Contains the price data of the reference index. Initial value: None
    :ivar pd.DataFrame index: Copy of `index` with computed indicators for signal generation. Initial value: None
    :ivar pd.DateRange dr: The range of dates between start and end date. Initial value: None
    :ivar pd.DateRange edr: Extended `dr` with additional data needed for calculating lagging indicators. Initial value: None
    :ivar bool dr_messages: If True, prints information on missing price data for `dr` and/or `edr` to the console. Initial value: None
    :ivar list trading_days: A list of all trading days between start and end date. Initial value: None
    :ivar list existing_symbols: All symbols included in the index during `dr` that exist in data\\lib. Initial value: None
    :ivar list missing_symbols: All symbols included in the index during `dr` that *do not* exist in data\\lib. Initial value: None
    :ivar pd.DataFrame eqc_df: A pd.DataFrame containing the equity curve of the backtest. Initial value: pd.DataFrame
    :ivar dict results: Contains high-level results (e.g. the winrate) of the backtest. Used to generate reports. Initial value: {}
    :ivar dict drawdown: Contains data used to calculate the (yearly) maximum drawdown. Initial value: {}
    :ivar dict exposure: Contains data on the net exposure to the market over time. Initial value: {}
    :ivar pd.DataFrame log_df: A pd.DataFrame containing a log of trades. Initial value: pd.DataFrame
    """

    def __init__(self, data_path):
        """Constructor Method. Connects the instance to the data folder. Fails if data folder not present.

        :param str data_path: The `data` directory or its parent directory.
        """

        self.data_path = connect(data_path, gspc_strict=True, hist_strict=True)
        self.data_path_symbols = [symbol[:-4] for symbol in os.listdir(self.data_path + 'lib\\') if '^' not in symbol]
        print(f"Setting the data_path to {data_path} was successful!")

        suffix = str(dt.datetime.now()).split('.')[0]
        suffix = suffix.replace(' ', '_')
        suffix = suffix.replace(':', '-')
        os.mkdir(self.data_path + 'backtest_' + suffix + '\\')
        self.output_path = self.data_path + 'backtest_' + suffix + '\\'

        self.input_data = {}
        self.data = None
        self.input_index = None
        self.index = None

        self.t0 = None
        self.runtime = None

        self.sd = None
        self.ed = None
        self.duration = None
        self.dr = None
        self.edr = None
        self.dr_messages = None
        self.trading_days = None

        self.constituents = None
        self.existing_symbols = None
        self.missing_symbols = None

        self.commission = None
        self.max_positions = None
        self.initial_equity = None
        self.max_trade_duration = None
        self.stoploss = None
        self.entry_mode = None

        self.signals = defaultdict(list)
        self.alerts = None
        self.eqc = {}
        self.eqc_df = None
        self.results = {}
        self.drawdown = {}
        self.exposure = {}
        self.correlations = {}
        self.log_df = pd.DataFrame()

        self.opt = False
        self.optimization_results = None
        self.best_parameters = None
        self.parameters = {}
        self.parameter_permutations = []

    @timeit_decorator
    def import_data(self, start_date, end_date, lag=200, date_range_messages=False):
        """Imports the necessary price data for the defined time period of the backtest.

        :param datetime.datetime or str start_date:
            The start date of the backtest. If weekend or US market closed, the next trading day will be set as start date.
        :param datetime.datetime or str end_date:
            The end date of the backtest. If weekend or US market closed, the next trading day will be set as end date.
        :param datetime.timedelta or int, default=200 lag:
            A timedelta that is added in front of the start_date. This is necessary for calculating indicators that
            depend on preceding price data such as moving averages among others.
        :param bool date_range_messages:
            If True, prints some extra information regarding missing price data for the backtest period to the console.

        :returns: None
        :rtype: NoneType
        """

        if not isinstance(start_date, dt.datetime):
            if not isinstance(start_date, str):
                raise TypeError("Parameter start_date must be of type datetime.datetime or str formatted 'YYYY-MM-DD'.")
            else:
                self.sd = dt.datetime.strptime(start_date, '%Y-%m-%d')
        else:
            self.sd = start_date

        if not isinstance(end_date, dt.datetime):
            if not isinstance(end_date, str):
                raise TypeError("Parameter start_date must be of type datetime.datetime or str formatted 'YYYY-MM-DD'.")
            else:
                self.ed = dt.datetime.strptime(end_date, '%Y-%m-%d')
        else:
            self.sd = end_date

        if not isinstance(lag, dt.timedelta):
            if not isinstance(lag, int):
                raise TypeError('Parameter lag must be of type datetime.timedelta or int.')
            else:
                lag = dt.timedelta(days=lag)

        self.duration = self.ed - self.sd
        self.dr = pd.date_range(self.sd, self.ed)
        self.edr = pd.date_range(self.sd - lag, self.ed)
        self.dr_messages = date_range_messages

        dtypes = {'symbol': str,
                  'open': np.float32,
                  'high': np.float32,
                  'low': np.float32,
                  'close': np.float32,
                  # 'volume': np.int64,
                  'dividends': np.float32,
                  'stock_splits': np.float32}
        self.input_index = pd.read_csv(self.data_path + 'lib\\^GSPC.csv', engine='c', dtype=dtypes,
                                       usecols=['date', 'symbol', 'open', 'high', 'low', 'close'],
                                       index_col='date', parse_dates=[0], dayfirst=True)
        self.input_index = self.input_index.loc[self.input_index.index.intersection(self.edr)]

        # self.alerts = "No alerts yet. Use the generate_signals or the scan_for_patterns method first."

        self.constituents = OrderedDict(sorted(constituents(self.sd, self.ed, lag, self.data_path).items()))

        self.existing_symbols = [s for s in self.constituents.keys() if s in self.data_path_symbols
                                 or ('*' in s and s[:-1] in self.data_path_symbols)]

        self.missing_symbols = [s for s in self.constituents.keys() if s not in self.existing_symbols]

        print(f'Constituents in time period: {len(self.constituents.keys())}. '
              f'Thereof not found in data path: {len(self.missing_symbols)}, {self.missing_symbols}')

        for symbol in self.existing_symbols:
            file = f"{self.data_path}lib\\{symbol[:-1] if '*' in symbol else symbol}.csv"
            dr, edr = self.constituents[symbol]['dr'], self.constituents[symbol]['edr']

            with open(file) as f:
                next(f)
                first = next(f).split(',')[0]
                first = dt.datetime.strptime(first, '%Y-%m-%d')

                skip = ((self.sd - first).days * 0.65).__floor__() - lag.days

            df = pd.read_csv(file, engine='c', dtype=dtypes, skiprows=range(1, skip), index_col='date',
                             parse_dates=['date'])
            df = df.loc[df.index.intersection(edr)]
            df['symbol'] = symbol

            if self.dr_messages:
                missing_records = df.index.symmetric_difference(edr.intersection(self.input_index.index))
                if any(missing_records):
                    print(f'{symbol}: {len(missing_records)} missing records in extended date range.')

            if not df.empty:
                self.input_data[symbol] = df

    @timeit_decorator
    def generate_signals(self, strategy):
        """Accepts an instance of the :class:`ndxtest.backtest.Strategy` class. Generates and oders the trading signals.

        :param ndxtest.backtest.Strategy strategy:
            A trading strategy built with the :class:`ndxtest.backtest.Strategy` class.

        :returns: None
        :rtype: NoneType
        """

        if not isinstance(strategy, Strategy):
            raise TypeError("Argument `strategy` must be an instance of ndxtest.backtest.Strategy.")

        strategy.data = self.input_data.copy()
        strategy.index = self.input_index.copy()
        self.data = strategy.generate_signals()

        for symbol, df in self.data.items():
            self.data[symbol] = df.loc[df.index.intersection(self.constituents[symbol]['dr'])]

        self.index = self.input_index.loc[self.input_index.index.intersection(self.dr)]
        self.index['d%change'] = self.index.close.pct_change() * 100
        self.index['c%change'] = ((self.index.close / self.index.close[0]) - 1) * 100
        self.index = self.index.round(2)

        self.trading_days = self.index.index.tolist()

        concat_data = pd.concat(self.data.values())
        concat_data.drop(columns=['high', 'low', 'close', 'volume', 'dividends', 'stock_splits'], inplace=True)
        dict_data = {}

        # self.alerts = concat_data.loc[(concat_data.index == self.trading_days[-1]) &
        #                               ((concat_data.entry_signals != 0) | (concat_data.exit_signals != 0))]

        for symbol, df in concat_data.groupby('symbol'):
            # trades will be executed on the next day
            df['enl_signals'] = df['enl_signals'].shift(1)
            df['exl_signals'] = df['exl_signals'].shift(1)
            df['ens_signals'] = df['ens_signals'].shift(1)
            df['exs_signals'] = df['exs_signals'].shift(1)

            # adding signals to exit on the last day
            df.loc[df.index[-1], ['score', 'enl_signals', 'exl_signals', 'ens_signals', 'exs_signals']] = \
                [0, 0, -2, 0, -2]  # check!

            # dropping NaN values
            df.dropna(inplace=True)
            # dropping emtpy rows
            dict_data[symbol] = df.loc[(df.enl_signals != 0) | (df.exl_signals != 0) | (df.ens_signals != 0) | (df.exs_signals != 0)]

        for date, values in pd.concat(dict_data.values()).groupby(level=0):
            # date, df: symbol, score, enl_signals, exl_signals, ens_signals, exs_signals

            self.signals[date] = values.to_dict(orient='records')
            self.signals[date] = sorted(list(self.signals[date]), key=lambda signal: signal['score'], reverse=True)

    @timeit_decorator
    def eval_signals(self, strategy):
        """This function searches the imported data for the signals specified by the strategy. It creates a .xlsx report
        about the immediate price action following the signals. It helps evaluating the viability of the signals
        generated by the strategy.

        :param ndxtest.backtest.Strategy strategy: A ndxtest.backtest.Strategy object.
        """

        if not isinstance(strategy, Strategy):
            raise TypeError("Argument `strategy` must be an instance of ndxtest.backtest.Strategy.")

        strategy_signals = {'entry_long_signals': strategy.condition_sets['enl'],
                            'exit_long_signals': strategy.condition_sets['exl'],
                            'entry_short_signals': strategy.condition_sets['ens'],
                            'exit_short_signals': strategy.condition_sets['exs']}

        with pd.ExcelWriter(self.output_path + 'signal_stats.xlsx') as writer:

            for signal_type, list_of_conditions in strategy_signals.items():
                results = {}

                for symbol, df in self.input_data.items():
                    signals = pd.DataFrame()

                    for i, condition in enumerate(list_of_conditions):
                        delta_days, func, index = condition
                        if not index:
                            signals[i] = func(df).shift(-delta_days)
                        else:
                            signals[i] = func(self.input_index).shift(-delta_days)

                    cds = signals.loc[signals.T.all()].index.values  # cds = completion days of pattern (=0)

                    results[symbol] = pd.DataFrame(data={'cd': cds,
                                                         'symbol': symbol,
                                                         'ep': df.shift(-1).loc[cds, 'open'],  # entry price
                                                         '+1c': df.shift(-1).loc[cds, 'close'],
                                                         '+1c_gt_ep': df.shift(-1).loc[cds, 'open'] < df.shift(-1).loc[
                                                             cds, 'close'],
                                                         '+2o': df.shift(-2).loc[cds, 'open'],
                                                         '+2o_gt_ep': df.shift(-1).loc[cds, 'open'] < df.shift(-2).loc[
                                                             cds, 'open'],
                                                         '+2c': df.shift(-2).loc[cds, 'close'],
                                                         '+2c_gt_ep': df.shift(-1).loc[cds, 'open'] < df.shift(-2).loc[
                                                             cds, 'close'],
                                                         '+3o': df.shift(-3).loc[cds, 'open'],
                                                         '+3o_gt_ep': df.shift(-1).loc[cds, 'open'] < df.shift(-3).loc[
                                                             cds, 'open'],
                                                         '+3c': df.shift(-3).loc[cds, 'close'],
                                                         '+3c_gt_ep': df.shift(-1).loc[cds, 'open'] < df.shift(-3).loc[
                                                             cds, 'close']
                                                         })

                d = pd.concat(results, ignore_index=True).sort_values(by='cd')

                if len(d.index) == 0:
                    print(f'No {signal_type} found!')
                    continue
                else:
                    d.dropna(inplace=True)

                    cols = [('+1c', '+1c_gt_ep'), ('+2o', '+2o_gt_ep'), ('+2c', '+2c_gt_ep'), ('+3o', '+3o_gt_ep'),
                            ('+3c', '+3c_gt_ep')]
                    stats = {'metric': ['Total count:',
                                        'Up count:',
                                        'Down count:',
                                        'Up %:',
                                        'Down %:',
                                        'Avg. % change:',
                                        'Med. % change:',
                                        'Std. %:',
                                        'Up avg. % change:',
                                        'Up med. % change:',
                                        'Up std. %:',
                                        'Up % change min:',
                                        'Up % change max:',
                                        'Down avg. % change:',
                                        'Down med. % change:',
                                        'Down std. %:',
                                        'Down % change min:',
                                        'Down % change max:',
                                        ]}

                    # making stats from the experimental data
                    for price_col, count_col in cols:
                        fc, tc = d[count_col].value_counts().sort_index().tolist()
                        pct_change = ((d[price_col] / d['ep']) - 1) * 100
                        t_pct_change = ((d.loc[d[count_col], price_col] / d.loc[d[count_col], 'ep']) - 1) * 100
                        f_pct_change = ((d.loc[~d[count_col], price_col] / d.loc[~d[count_col], 'ep']) - 1) * 100
                        stats[count_col] = [tc + fc,
                                            tc,
                                            fc,
                                            round((tc / (fc + tc) * 100), 2),
                                            round((fc / (fc + tc) * 100), 2),
                                            round(pct_change.mean(), 2),
                                            round(pct_change.median(), 2),
                                            round(pct_change.std(ddof=1), 2),
                                            round(t_pct_change.mean(), 2),
                                            round(t_pct_change.median(), 2),
                                            round(t_pct_change.std(ddof=1), 2),
                                            round(t_pct_change.min(), 2),
                                            round(t_pct_change.max(), 2),
                                            round(f_pct_change.mean(), 2),
                                            round(f_pct_change.median(), 2),
                                            round(f_pct_change.std(ddof=1), 2),
                                            round(f_pct_change.min(), 2),
                                            round(f_pct_change.max(), 2)
                                            ]

                    d.to_excel(writer, sheet_name=signal_type, index=False)
                    pd.DataFrame(data=stats).to_excel(writer, sheet_name=signal_type + '_stats', index=False)

    @timeit_decorator
    def run_backtest(self, commission=.001, max_positions=10, initial_equity=10000.00,
                     max_trade_duration=None, stoploss=None, detailed_eqc=True):
        """Executes the backtest and creates several logs in the meantime.

        :param float, default=.001 commission:
            Commission that is paid upon entering/exiting positions.
        :param int, default=10 max_positions:
            Maximum number of positions (slots) in the portfolio.
        :param float, deault=10000.00 initial_equity:
            The initial capital to start with.
        :param int, default=None max_trade_duration:
            Maximum number of days before positions will be closed. Will close positions independent of any signals provided by the strategy.
        :param float, deaulft=None stoploss:
            The maximum % (e.g. 0.05) of adverse price movement before positions will be closed. Will close positions indepent of signals provided by the strategy.
        :param bool, default=True detailed_eqc:
            If True, calculates the ohlc market value of the portfolio daily during the backtest period. Significantly increases time for computation of backtest. Set this to false for quick testing.

        :returns: None
        :rtype: None
        """

        if self.output_path is None:
            suffix = str(dt.datetime.now()).split('.')[0]
            suffix = suffix.replace(' ', '_')
            suffix = suffix.replace(':', '-')
            self.output_path = self.data_path + suffix + '\\'

        self.commission = commission
        self.max_positions = max_positions
        self.initial_equity = initial_equity
        self.max_trade_duration = max_trade_duration
        self.stoploss = stoploss
        self.exposure = {}

        p = Portfolio(max_positions=max_positions, initial_equity=initial_equity, commission=commission)

        for date in self.trading_days:
            max_trade_duration_signals = []
            stoploss_signals = []

            if max_trade_duration is not None and p.positions:
                current_positions = list(p.long_positions.values()) + list(p.short_positions.values())

                max_trade_duration_violated = \
                    list(filter(lambda position: position['entry_date'] <= date - dt.timedelta(days=max_trade_duration),
                                current_positions))

                max_trade_duration_signals = [{'symbol': position['symbol'],
                                               'enl_signals': 0,
                                               'exl_signals': -2,
                                               'ens_signals': 0,
                                               'exs_signals': -2,
                                               'score': 0,
                                               'open': self.data[position['symbol']].loc[date, 'open']}
                                              for position in max_trade_duration_violated]

            if stoploss is not None and p.positions:
                long_positions, short_positions = list(p.long_positions.values()), list(p.short_positions.values())

                long_stoploss_violated = list(filter(lambda position:
                                                     position['entry_price'] * (1 - stoploss) >
                                                     self.data[position['symbol']].loc[date, 'open'], long_positions))

                short_stoploss_violated = list(filter(lambda position:
                                                      position['entry_price'] * (1 + stoploss) <
                                                      self.data[position['symbol']].loc[date, 'open'], short_positions))

                stoploss_signals = [{'symbol': position['symbol'],
                                     'enl_signals': 0,
                                     'exl_signals': -2,
                                     'ens_signals': 0,
                                     'exs_signals': -2,
                                     'score': 0,
                                     'open': self.data[position['symbol']].loc[date, 'open']}
                                    for position in long_stoploss_violated + short_stoploss_violated]

            self.signals[date] += max_trade_duration_signals
            self.signals[date] += stoploss_signals

            # processing exit signals
            if self.signals[date]:

                for s in [s for s in self.signals[date] if s['symbol'] in p.positions()]:

                    # exiting short position
                    if s['exs_signals'] in {1, -2} and s['symbol'] in p.short_positions:
                        p.long(data={'symbol': s['symbol'],
                                     'exit_score': s['score'],
                                     'exit_price': s['open'],
                                     'exit_date': date})

                    # exiting long position
                    if s['exl_signals'] in {1, -2} and s['symbol'] in p.long_positions:
                        p.short(data={'symbol': s['symbol'],
                                      'exit_score': s['score'],
                                      'exit_price': s['open'],
                                      'exit_date': date})

                if date == self.trading_days[-1]:
                    self.signals[date] = []
                else:
                    # all symbols with existing positions are already filtered out here
                    self.signals[date] = list(filter(lambda signal: signal['symbol'] not in p.positions(), self.signals[date]))

                # processing entry signals
                for s in self.signals[date]:

                    # entering long
                    if s['enl_signals'] == 1 and s['symbol'] and p.free_slot():
                        s['nshares'] = p.calculate_nshares(s['open'])
                        if s['nshares'] > 0:
                            p.long(data={'symbol': s['symbol'],
                                         'signal': 1,
                                         'entry_score': s['score'],
                                         'entry_price': s['open'],
                                         'nshares': s['nshares'],
                                         'entry_date': date})
                            continue  # if enl and ens signals are present enl has precedence

                    # entering short
                    if s['ens_signals'] == 1 and s['symbol'] and p.free_slot():
                        s['nshares'] = -1 * p.calculate_nshares(s['open'])
                        if s['nshares'] < 0:
                            p.short(data={'symbol': s['symbol'],
                                          'signal': 1,
                                          'entry_score': s['score'],
                                          'entry_price': s['open'],
                                          'nshares': s['nshares'],
                                          'entry_date': date})

            # logging the "market exposure" statistic
            self.exposure[date] = p.current_exposure()

            if detailed_eqc:
                op, hi, lo, cl, data = 0, 0, 0, 0, None
                if p.positions:
                    for symbol in p.positions():
                        if symbol in p.long_positions.keys():
                            nshares = p.long_positions[symbol]['nshares']
                        else:
                            nshares = p.short_positions[symbol]['nshares']
                        long_positions = self.data[symbol].loc[date, ['open', 'high', 'low', 'close']]
                        op += long_positions.open * nshares
                        hi += long_positions.high * nshares
                        lo += long_positions.low * nshares
                        cl += long_positions.close * nshares
                data = {'open': p.cash + p.restricted_cash() + op, 'high': p.cash + p.restricted_cash() + hi,
                        'low': p.cash + p.restricted_cash() + lo, 'close': p.cash + p.restricted_cash() + cl}
                self.eqc[date] = data

        self.exposure = pd.DataFrame.from_dict(data=self.exposure, orient='index', dtype=np.int8)

        self.log_df = p.create_log_df()
        self.log_df.to_csv(self.output_path + 'tradelog.csv')
        self.log_df = self.log_df.loc[self.log_df['exit_date'] != 0]

        if detailed_eqc:
            self.eqc_df = pd.DataFrame.from_dict(data=self.eqc, orient='index')
            self.eqc_df['d%change'] = self.eqc_df.close.pct_change() * 100
            self.eqc_df['c%change'] = (((self.eqc_df.close / initial_equity) - 1) * 100)
            self.eqc_df = self.eqc_df.round(2)
            self.eqc_df.to_csv(self.output_path + 'equity_curve.csv')
        else:
            x = self.log_df.drop_duplicates(subset=['exit_date'], keep='last')
            x = x.set_index('exit_date')['market_value']
            self.eqc_df = pd.DataFrame(data={'close': np.nan}, index=self.index.index)
            self.eqc_df.loc[x.index, 'close'] = x
            self.eqc_df.fillna(method="ffill", inplace=True)
            self.eqc_df.fillna(self.initial_equity, inplace=True)
            self.eqc_df['d%change'] = self.eqc_df.close.pct_change() * 100
            self.eqc_df['c%change'] = (((self.eqc_df.close / initial_equity) - 1) * 100)
            self.eqc_df = self.eqc_df.round(2)
            self.eqc_df.to_csv(self.output_path + 'equity_curve.csv')

        self.drawdown = pd.DataFrame(data={'roll_max': self.eqc_df.close.rolling(252, min_periods=1).max()},
                                     index=self.eqc_df.index)
        self.drawdown['daily'] = (self.eqc_df.close / self.drawdown.roll_max - 1) * 100
        self.drawdown['daily_max'] = self.drawdown.daily.rolling(252, min_periods=1).min()
        self.drawdown = self.drawdown.round(2)

        self.results = {'n_trades': p.number_of_trades,
                        'max_drawdown': self.drawdown.daily_max.min(),
                        'median%pnl': self.log_df['%change'].median().__round__(2),
                        'min%pnl': self.log_df['%change'].min(),
                        'max%pnl': self.log_df['%change'].max(),
                        'median_dur': self.log_df.duration.median(),
                        'min_dur': self.log_df.duration.min(),
                        'max_dur': self.log_df.duration.max(),
                        '%profitable': (np.count_nonzero(self.log_df.profitable != 0) / len(
                            self.log_df.profitable) * 100).__round__(2),
                        'commission_paid': p.commission_paid.__round__(2),
                        'perf': self.eqc_df['c%change'][-1],
                        'ann_perf': (((1 + ((self.eqc_df.close[-1] - initial_equity) / initial_equity))
                                      ** (1 / (self.duration.days / 365.25)) - 1) * 100).__round__(2),
                        'bm_perf': self.index['c%change'][-1],
                        'bm_ann_perf': (((1 + (
                                (self.index.close[-1] - self.index.close[0]) / self.index.close[0]))
                                         ** (1 / (self.duration.days / 365.25)) - 1) * 100).__round__(2)}

    @timeit_decorator
    def report(self):
        """Generates .xlsx and .pdf reports of the last run backtest.

        :returns: None
        :rtype: NoneType
        """

        fig1, axs1 = plt.subplots(3, 1, figsize=(9, 6))
        axs1[0].plot(self.index['c%change'].index, self.index['c%change'], color='blue', label='Benchmark')
        axs1[0].plot(self.eqc_df['c%change'].index, self.eqc_df['c%change'], color='black',
                     label='Backtest')
        axs1[0].set_xticks([])
        axs1[0].set_xticklabels([])
        axs1[0].set_ylabel('Cumulative % Change')
        axs1[0].grid(axis='y')
        axs1[0].legend()
        axs1[1].plot(self.drawdown['daily'].index, self.drawdown['daily'], color='black', label='Drawdown')
        axs1[1].plot(self.drawdown['daily_max'].index, self.drawdown['daily_max'], color='red',
                     label='Max. 1Yr Drawdown')
        axs1[1].set_xticks([])
        axs1[1].set_xticklabels([])
        axs1[1].set_ylabel('1Yr Drawdown')
        axs1[1].grid(axis='y')
        axs1[1].legend()
        axs1[2].plot(self.exposure['n_long'].index, self.exposure['n_long'], color='blue', label='Long Positions')
        axs1[2].plot(self.exposure['n_short'].index, self.exposure['n_short'], color='red', label='Short Positions')
        axs1[2].plot(self.exposure['n_free'].index, self.exposure['n_free'], color='grey', label='Free Slots')
        axs1[2].set_xlabel('Date')
        axs1[2].set_ylabel('Positions')
        axs1[2].grid(axis='y')
        axs1[2].legend()
        fig1.tight_layout()

        plt.savefig(self.output_path + "f1.png", dpi=None, facecolor='w', edgecolor='w')

        class PDF(FPDF):  # A4: w = 210, h = 297
            pass

        pdf = PDF()
        pdf.set_left_margin(margin=30)
        pdf.add_page()
        pdf.set_font('helvetica', size=12)

        pdf.cell(w=80, align='', txt=f'BackTest Report:', ln=0)
        pdf.cell(w=80, align='', txt=f'{dt.datetime.now().date()}', ln=1)
        pdf.cell(w=80, align='', txt=f'Benchmark Index:', ln=0)
        pdf.cell(w=80, align='', txt=f'SPY', ln=1)

        pdf.ln(140)
        pdf.image(self.output_path + 'f1.png', x=15, y=25, w=180, h=120)

        # pdf.cell(w=80, align='', txt=f'Parameters:', ln=0)
        # pdf.cell(w=80, align='', txt=f'{self.best_parameters if self.opt else "not optimized"}', ln=1)
        pdf.cell(w=80, align='', txt=f'Start date:', ln=0)
        pdf.cell(w=80, align='', txt=f'{self.sd.date()}', ln=1)
        pdf.cell(w=80, align='', txt=f'End date:', ln=0)  #
        pdf.cell(w=80, align='', txt=f'{self.ed.date()}', ln=1)
        pdf.cell(w=80, align='', txt=f"Initial Equity [$]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.initial_equity}", ln=1)
        pdf.cell(w=80, align='', txt=f"Final Equity [$]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.eqc_df.close[-1]}", ln=1)
        pdf.cell(w=80, align='', txt=f"Max. Drawdown [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['max_drawdown']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Performance Strategy [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['perf']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Performance Benchmark [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{str(self.results['bm_perf'])[:5]}", ln=1)
        pdf.cell(w=80, align='', txt=f"Ann. Performance Strategy [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['ann_perf']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Ann. Performance Benchmark [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['bm_ann_perf']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Total commissions paid [$]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['commission_paid']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Number of trades executed:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['n_trades']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Median trade duration [Days]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['median_dur']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Worst trade [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['min%pnl']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Best trade [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['max%pnl']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Median trade [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['median%pnl']}", ln=1)
        pdf.cell(w=80, align='', txt=f"Profitable trades [%]:", ln=0)
        pdf.cell(w=80, align='', txt=f"{self.results['%profitable']}", ln=1)

        pdf.output(self.output_path + 'backtest_report.pdf', 'F')
        os.remove(self.output_path + 'f1.png')

    def query_missing_records(self, date):
        """Queries, whether any symbols lack price data for a specific date. (for debugging the library)

        :param datetime.datetime or str date: If string it must be 'YYYY-MM-DD' formatted.

        :returns: None
        :rtype: NoneType
        """

        if not isinstance(date, dt.datetime):
            if isinstance(date, str):
                date = dt.datetime.strptime(date, '%Y-%m-%d')
            else:
                raise TypeError("Please provide a 'YYYY-MM-DD' formatted string or a datetime.datetime object.")

        if date.isoweekday() in {6, 7}:
            raise ValueError(f'{date} is not a weekday. Call function with weekdays only.')

        if date < self.sd:
            print(f'Warning: {date} is not within the backtest period. The record might be available but not imported.')

        for k, v in self.input_data.items():
            if date not in v.index:
                if date in self.constituents[k]['dr']:
                    print(f'Data for {k} is missing the record for {date}.')



""""""
import numpy as np
import pandas as pd

# General Indicators #


def roc(arr, n, input_col='close'):
    """Calculates the rate of change (ROC).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int n: The periods for ROC calculation.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`

    :returns: A pd.Series containing the ROC. Will contain np.nan values.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    return arr.pct_change(periods=n)


def sma(arr, n, input_col='close', full_window=True):
    """Calculates the simple moving average (SMA).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int n: The periods for SMA calculation.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the SMA. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    return arr.rolling(n, min_periods=n if full_window else 1).mean()


def ema(arr, n, input_col='close', full_window=True):
    """Calculates the exponential moving average (EMA).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int n: The periods for EMA calculation.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the EMA. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    return arr.ewm(span=n, min_periods=n if full_window else 1).mean()


def ssma(arr, n, input_col='close', full_window=True):
    """Calculates smoothed simple moving average (SSMA).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int n: The periods for SSMA calculation.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the SSMA. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    return arr.ewm(alpha=1 / n, min_periods=n if full_window else 1).mean()


def atr(arr: pd.DataFrame, n=14, full_window=True):
    """Calculates the average true range (ATR) indicator.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.
    :param int, default=14 n: The periods for the atr calculation. Usually the 14 period atr is used.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the ATR. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        true_range = np.maximum.reduce([arr.high, arr.close.shift(1)]) - np.minimum.reduce(
            [arr.low, arr.close.shift(1)])
        true_range = pd.Series(data=true_range, index=arr.index)
        average_true_range = true_range.ewm(alpha=1 / n, min_periods=n if full_window else 1).mean()
        # relative_true_range = (true_range / average_true_range).astype(_DTYPE)
        return average_true_range
    else:
        raise TypeError("Argument 'arr' must be a pd.DataFrame.")


def adx(arr: pd.DataFrame, n=14, full_window=True):
    """Calculates the average directional index (ADX) indicator. Part of calculating the ADX is calculating the ATR first.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.
    :param int, default=14 n: The periods for ADX calculation. Usually the 14 period ADX is used.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the ADX. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        plus_dm = (arr.high - arr.high.shift(1))
        minus_dm = (arr.low.shift(1) - arr.low)
        positive_dm = np.where(np.greater(plus_dm, minus_dm) & np.greater(plus_dm, 0), plus_dm, 0)
        positive_dm = pd.Series(positive_dm, index=arr.index)
        negative_dm = np.where(np.greater(minus_dm, plus_dm) & np.greater(minus_dm, 0), minus_dm, 0)
        negative_dm = pd.Series(negative_dm, index=arr.index)
        average_true_range = atr(arr, n, full_window=full_window)

        positive_di = (ema(positive_dm, n, full_window=full_window) / average_true_range) * 100
        negative_di = (ema(negative_dm, n, full_window=full_window) / average_true_range) * 100
        directional_index = (abs(positive_di - negative_di) / (positive_di + negative_di)) * 100
        average_directional_index = \
            directional_index.ewm(alpha=1 / n, min_periods=n if full_window else 1).mean()
        return average_directional_index
    else:
        raise TypeError("Argument 'arr' must be a pd.DataFrame.")


def rsi(arr, n=14, input_col='close', full_window=True):
    """Calculates the relative strength index (RSI).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int, default=14 n: The periods for the RSI. By convention the 14 period RSI is calculated.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: A pd.Series containing the RSI. Will contain np.nan values if full_window=True.
    :rtype: pd.Series of dtype np.float32.
    """

    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]

    moves = arr.diff()
    up_moves = pd.Series(data=np.where(np.greater_equal(moves, 0), moves, 0), index=moves.index)
    down_moves = abs(pd.Series(data=np.where(np.greater_equal(0, moves), moves, 0), index=moves.index))
    rs_index = 100 - (100 / (1 + (up_moves.rolling(n, min_periods=n if full_window else 1).mean() /
                                  down_moves.rolling(n, min_periods=n if full_window else 1).mean())))
    return rs_index


def macd(arr, input_col='close', full_window=True):
    """Calculates the moving average convergence divergence (MACD) index. The MACD has fixed periods (12 and 26).

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`.
    :param bool, default=True full_window: If True, the output pd.Series will have 25 trailing np.nan values.

    :returns: Three separate pd.Series: The MACD line, the signal line and the MACD histogram.
    :rtype: a tuple containing three pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    macd_line = arr.ewm(span=12, min_periods=12 if full_window else 1).mean() - \
                arr.ewm(span=26, min_periods=26 if full_window else 1).mean()
    signal_line = macd_line.ewm(span=9, min_periods=9 if full_window else 1).mean()
    hist = macd_line - signal_line
    return macd_line, signal_line, hist


def bbands(arr, n=20, stdl=2, stdu=2, input_col="close", full_window=True):
    """Calculates the a simple moving average (SMA) and its bollinger bands (BB). Usually, the 20 period SMA and its BB's are used.

    :param pd.DataFrame or pd.Series arr: A pd.DataFrame containing ohlc price data, or a pd.Series containing price data.
    :param int n: The periods for SMA calculation.
    :param float stdl: The standard deviation used in calculating the *lower* BB.
    :param float stdu: The standard deviation used in calculating the *upper* BB.
    :param str, default='close' input_col: The input column name if arr is a `pd.DataFrame`.
    :param bool, default=True full_window: If True, the output pd.Series will have n-1 trailing np.nan values.

    :returns: Three separate pd.Series: The SMA, the lower BB and the upper BB.
    :rtype: a tuple containing three pd.Series of dtype np.float32.
    """
    if isinstance(arr, pd.DataFrame):
        arr = arr[input_col]
    ma = arr.rolling(n, min_periods=n if full_window else 1).mean()
    lbb = (ma - arr.rolling(n, min_periods=n if full_window else 1).std() * stdl)
    ubb = (ma + arr.rolling(n, min_periods=n if full_window else 1).std() * stdu)
    return ma, lbb, ubb


# Crossovers #


def static(s, arr):
    """Generates a static level to be used in crossover strategies.

    :param int s: An integer.
    :param pd.DataFrame or pd.Series arr: A pd.DataFrame or pd.Series.

    :returns: A pd.Series that has the index of arr and values of s.
    :rtype: pd.Series of dtype np.int8
    """
    return pd.Series(data=s, index=arr.index, dtype=np.int8)


def crossover(slow: pd.Series or int or float, fast: pd.Series, direction: str):
    """Checks for bullish or bearish crossovers of a faster-moving series through a slower-moving or static series.

    The parameter `slow` can be either a pd.Series or an integer. If an integer or float is provided it will be converted to
    a static pd.Series with the index of `fast`.

    Examples of usage:

    - crossover(sma(arr, 50), sma(arr, 20), 'bullish') -> Bullish crossovers of the 20 period SMA through the 50 period SMA.
    - crossover(50, rsi(arr), 'bearish') -> Bearish crossovers whenever the RSI falls below the 50 level.

    :param pd.Series or int or float slow: A pd.Series, an integer or a float.
    :param pd.Series fast: A pd.Series.
    :param str, direction: Must be either 'bullish' or 'bearish'.

    :returns: A pd.Series. True where bullish or bearish crossovers have occurred, respectively.
    :rtype: pd.Series of dtype bool
    """

    if not isinstance(fast, pd.Series):
        raise TypeError("Argument 'fast' must be of type pd.Series.")

    if not isinstance(slow, pd.Series):
        if isinstance(slow, int) or isinstance(slow, float):
            slow = pd.Series(data=slow, index=fast.index)
        else:
            raise TypeError("Argument 'slow' must be of type pd.Series or int.")

    if not isinstance(direction, str):
        raise TypeError("Argument 'direction' must be of type str.")

    if direction == 'bullish':
        bullish = True
    elif direction == 'bearish':
        bullish = False
    else:
        raise ValueError("Argument 'direction' must be either 'bullish' or 'bearish'.")

    bearish_crossovers = ((slow.shift(1) < fast.shift(1)) & (fast < slow))
    bullish_crossovers = ((slow.shift(1) > fast.shift(1)) & (fast > slow))
    return bullish_crossovers if bullish else bearish_crossovers


def intraday_price_crossover(arr: pd.DataFrame, series: pd.Series, direction: str):
    """Checks for bullish or bearish intraday-crossovers of the price through a moving series.

    Examples:

    - intraday_price_crossover(arr, sma(arr, 100), 'bearish') -> True where a candle opens above and closes beneath the 100 period SMA.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.
    :param pd.Series series: A pd.Series, e.g. a moving average.
    :param str, direction: Must be either 'bullish' or 'bearish'.

    :returns: A pd.Series. True where bullish or bearish crossovers have occurred, respectively.
    :rtype: pd.Series of dtype bool
    """

    if not isinstance(arr, pd.DataFrame):
        raise TypeError("Argument 'arr' must be of type pd.DataFrame.")

    if not isinstance(series, pd.Series):
        raise TypeError("Argument 'series' must be of type pd.Series.")

    if not isinstance(direction, str):
        raise TypeError("Argument 'direction' must be of type str.")

    if direction == 'bullish':
        bullish = True
    elif direction == 'bearish':
        bullish = False
    else:
        raise ValueError("Argument 'direction' must be either 'bullish' or 'bearish'.")

    bullish_crossovers = ((series > arr.open) & (arr.close > series))
    bearish_crossovers = ((series < arr.open) & (arr.close < series))

    return bullish_crossovers if bullish else bearish_crossovers


# Candlesticks and Price Action #


def prev(arr):
    """Simply shifts `arr` by 1. Provides syntactic sugar.

    :param pd.Series or pd.DataFrame arr: A pd.Series or ps.DataFrame.

    :returns: A pd.Series or pd.DataFrame, shifted by 1.
    :rtype: pd.Series or pd.DataFrame
    """

    if not isinstance(arr, pd.Series) or isinstance(arr, pd.DataFrame):
        raise TypeError("Parameter `arr` must be of type `pd.Series` or `pd.DataFrame`.")

    return arr.shift(1)


def green_candle(arr):
    """Green candle.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.

    :returns: A pd.Series, True for all candles that closed higher than they opened.
    :rtype: pd.Series of dtype bool
    """

    return arr['close'] > arr['open']


def red_candle(arr):
    """Red candle.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.

    :returns: A pd.Series, True for all candles that closed lower than they opened.
    :rtype: pd.Series of dtype bool
    """

    return arr['close'] < arr['open']


def bullish_pin_bar(arr, j=3):
    """Detects bullish pin bars.

    Herein, bullish pin bars are defined as follows:

    - The lower wick of the candle must be at least three (j) times longer than candle body.
    - The lower boundary of the candle-body must be situated in the upper third of the candle.
    - The lower wick of the pin bar must reach a lower low than the previous candle.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.
    :param int, default=3 j: The ratio of lower wick length to candle body length.

    :returns: A pd.Series, True where bullish pin bars were detected.
    :rtype: pd.Series of dtype bool
    """

    body_length = abs(arr['close'] - arr['open'])
    lower_wick_length = np.minimum(arr['close'], arr['open']) - arr['low']
    # prev_min = np.minimum(arr['close'].shift(1), arr['open'].shift(1))
    # prev_max = np.maximum(arr['close'].shift(1), arr['open'].shift(1))
    # body_inside_prev_body = (prev_min < arr['open']) & (arr['open'] < prev_max) & \
    #                           (prev_min < arr['close']) & (arr['close'] < prev_max)
    body_in_upper_third = np.minimum(arr['open'], arr['close']) > (arr['low'] + (2 * arr['high'])) / 3
    wick_lt_prev = arr['low'] < arr['low'].shift(1)
    return (lower_wick_length > j * body_length) & body_in_upper_third & wick_lt_prev


def bearish_pin_bar(arr, j=3):
    """Detects bearish pin bars.

    Herein, bearish pin bars are defined as follows:

    - The upper wick of the candle must be at least three (j) times longer than candle body.
    - The upper boundary of the candle-body must be situated in the lower third of the candle.
    - The upper wick of the pin bar must reach a higher high than the previous candle.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.
    :param int, default=3 j: The ratio of lower wick length to candle body length.

    :returns: A pd.Series, True where bearish pin bars were detected.
    :rtype: pd.Series of dtype bool
    """

    body_length = abs(arr['close'] - arr['open'])
    upper_wick_length = arr['high'] - np.maximum(arr['close'], arr['open'])
    # prev_min = np.minimum(arr['close'].shift(1), arr['open'].shift(1))
    # prev_max = np.maximum(arr['close'].shift(1), arr['open'].shift(1))
    # body_inside_prev_body = (prev_min < arr['open']) & (arr['open'] < prev_max) & \
    #                          (prev_min < arr['close']) & (arr['close'] < prev_max)
    body_in_lower_third = np.maximum(arr['open'], arr['close']) < ((2 * arr['low']) + arr['high']) / 3
    wick_gt_prev = (arr['high'] < arr['high'].shift(1)) & (arr['low'] < arr['low'].shift(-1))
    return (upper_wick_length > j * body_length) & body_in_lower_third & wick_gt_prev


def inside_bar(arr):
    """Detects inside bars.

    Inside bars are defined as follows:

    - The high of the candle must be lower than the previous candles high.
    - The low of the candle must be higher than the previous candles low.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.

    :returns: A pd.Series, True where inside bars were detected.
    :rtype: pd.Series of dtype bool
    """
    return (arr['high'] < arr['high'].shift(1)) & (arr['low'] > arr['low'].shift(1))


def inside_open(arr):
    """Detects candles, opening within the body of the previous candle.

    :param pd.DataFrame arr: A pd.DataFrame containing ohlc price data.

    :returns: A pd.Series, True where the candles open was within the previous candles body.
    :rtype: pd.Series of dtype bool
    """

    return (np.minimum(arr['close'].shift(1), arr['open'].shift(1)) < arr['open']) & \
           (np.maximum(arr['close'].shift(1), arr['open'].shift(1)) > arr['open'])



"""This is ndxtest. It lets you test trading strategies on the S&P 500 index.

Refer to https://github.com/lksdnx/ndxtest and https://ndxtest.readthedocs.io/en/latest/
for information on how to use this package.

For information on the usage of this package please refer to the online documentation.

Modules
_______
    backtest.py
    utils.py
    indicators.py
"""

import backtest
import indicators
import utils




from setuptools import setup, find_packages

# del build -Recurse; del dist -Recurse; del ndxtest.egg-info -Recurse
# py setup.py sdist bdist_wheel

# uploading to test pypi
# twine upload --repository-url https://test.pypi.org/legacy/ dist/*

# installing from test pypi
# pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple ndxtest

# uploading to pypi
# twine upload dist/*


with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["numpy>=1.21.0",
                "pandas>=1.3.0",
                "yfinance>=0.1.74",
                "matplotlib>=3.4.2",
                "fpdf>=1.7.2",
                "openpyxl>=3.0.7"]

# deleted sphinx from here

setup(name="ndxtest", version="0.0.1",
      author="lksdnx",
      author_email="lukas.dnx@gmail.com",
      description="ndxtest lets you test trading strategies on the S&P 500.",
      long_description=readme,
      long_description_content_type="text/markdown",
      url="https://github.com/lksndx/ndxtest/",
      packages=find_packages(),
      install_requires=requirements,
      classifiers=["Programming Language :: Python :: 3.9", "License :: OSI Approved :: MIT License"]
      )
